"""Module for extending python-docx Paragraph objects."""

import bisect
import dataclasses
import itertools
import re

from docx.enum import text
from docx.text import paragraph as docx_paragraph

from cmi_docx import run


@dataclasses.dataclass
class FindParagraph:
    """Data class for maintaing find results in paragraphs.

    Attributes:
        paragraph: The paragraph containing the text.
        character_indices: A list of matching character indices of the text in the
            paragraph.
    """

    paragraph: docx_paragraph.Paragraph
    character_indices: list[tuple[int, int]]


class ExtendParagraph:
    """Extends a python-docx Word paragraph with additional functionality."""

    def __init__(self, paragraph: docx_paragraph.Paragraph) -> None:
        """Initializes an ExtendParagraph object.

        Args:
            paragraph: The paragraph to extend.
        """
        self.paragraph = paragraph

    def find_in_paragraph(self, needle: str) -> FindParagraph:
        """Finds the indices of a text relative to the paragraph.

        Args:
            needle: The text to find.

        Returns:
            The indices of the text in the paragraph.
        """
        within_paragraph_indices = [
            (match.start(), match.end())
            for match in re.finditer(re.escape(needle), self.paragraph.text)
        ]

        return FindParagraph(
            paragraph=self.paragraph,
            character_indices=within_paragraph_indices,
        )

    def find_in_runs(self, needle: str) -> list[run.FindRun]:
        """Finds the indices of a text relative to the paragraph's runs.

        Args:
            needle: The text to find.

        Returns:
            The indices of the text in the paragraph.
        """
        run_finds: list[run.FindRun] = []
        run_lengths = [len(run.text) for run in self.paragraph.runs]
        cumulative_run_lengths = list(itertools.accumulate(run_lengths))
        for occurence in self.find_in_paragraph(needle).character_indices:
            start_run = bisect.bisect_right(cumulative_run_lengths, occurence[0])
            end_run = bisect.bisect_right(
                cumulative_run_lengths, occurence[1], lo=start_run
            )
            start_index = (
                occurence[0] - cumulative_run_lengths[start_run - 1]
                if start_run > 0
                else occurence[0]
            )
            end_index = (
                occurence[1] - cumulative_run_lengths[end_run - 1]
                if end_run > 0
                else occurence[1]
            )

            run_finds.append(
                run.FindRun(
                    paragraph=self.paragraph,
                    run_indices=(start_run, end_run),
                    character_indices=(start_index, end_index),
                )
            )
        return run_finds

    def replace(self, needle: str, replace: str) -> None:
        """Finds and replaces text in a Word paragraph.

        Args:
            needle: The text to find.
            replace: The text to replace.
        """
        run_finder = self.find_in_runs(needle)
        run_finder.sort(
            key=lambda x: (x.run_indices[0], x.character_indices[0]), reverse=True
        )

        for run_find in run_finder:
            run_find.replace(replace)

    def format(
        self,
        *,
        bold: bool | None = None,
        italics: bool | None = None,
        font_size: int | None = None,
        font_rgb: tuple[int, int, int] | None = None,
        line_spacing: float | None = None,
        space_before: float | None = None,
        space_after: float | None = None,
        alignment: text.WD_PARAGRAPH_ALIGNMENT | None = None,
    ) -> None:
        """Formats a paragraph in a Word document.

        Args:
            bold: Whether to bold the paragraph.
            italics: Whether to italicize the paragraph.
            font_size: The font size of the paragraph.
            font_rgb: The font color of the paragraph.
            line_spacing: The line spacing of the paragraph.
            space_before: The spacing before the paragraph.
            space_after: The spacing after the paragraph.
            alignment: The alignment of the paragraph.
        """
        if line_spacing is not None:
            self.paragraph.paragraph_format.line_spacing = line_spacing

        if alignment is not None:
            self.paragraph.alignment = alignment

        if space_before is not None:
            self.paragraph.paragraph_format.space_before = space_before

        if space_after is not None:
            self.paragraph.paragraph_format.space_after = space_after

        for paragraph_run in self.paragraph.runs:
            run.ExtendRun(paragraph_run).format(
                bold=bold,
                italics=italics,
                font_size=font_size,
                font_rgb=font_rgb,
            )



"""Utility functions for the cmi_docx package."""


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """Converts RGB values to a hexadecimal color code.

    Args:
        r: The red component of the RGB color.
        g: The green component of the RGB color.
        b: The blue component of the RGB color.

    Returns:
        The hexadecimal color code representing the RGB color.
    """
    return f"#{r:02x}{g:02x}{b:02x}".upper()



"""Extends a python-docx Table cell with additional functionality."""

from docx import oxml, table
from docx.enum import text
from docx.oxml import ns

from cmi_docx import paragraph, utils


class ExtendCell:
    """Extends a python-docx Word cell with additional functionality."""

    def __init__(self, cell: table._Cell) -> None:
        """Initializes an ExtendCell object.

        Args:
            cell: The cell to extend.
        """
        self.cell = cell

    def format(
        self,
        *,
        line_spacing: float | None = None,
        space_before: float | None = None,
        space_after: float | None = None,
        bold: bool | None = None,
        italics: bool | None = None,
        font_size: int | None = None,
        font_rgb: tuple[int, int, int] | None = None,
        background_rgb: tuple[int, int, int] | None = None,
        alignment: text.WD_PARAGRAPH_ALIGNMENT | None = None,
    ) -> None:
        """Formats a cell in a Word table.

        Args:
            cell: The cell to format.
            line_spacing: The line spacing of the cell.
            space_before: The spacing before the cell.
            space_after: The spacing after the cell.
            bold: Whether to bold the cell.
            italics: Whether to italicize the cell.
            font_size: The font size of the cell.
            font_rgb: The font color of the cell.
            background_rgb: The background color of the cell.
            alignment: The alignment of the cell.
        """
        for table_paragraph in self.cell.paragraphs:
            paragraph.ExtendParagraph(table_paragraph).format(
                line_spacing=line_spacing,
                bold=bold,
                italics=italics,
                font_size=font_size,
                font_rgb=font_rgb,
                alignment=alignment,
                space_after=space_after,
                space_before=space_before,
            )

        if background_rgb is not None:
            shading = oxml.parse_xml(
                (
                    r'<w:shd {} w:fill="'
                    + f"{utils.rgb_to_hex(*background_rgb)}"
                    + r'"/>'
                ).format(
                    ns.nsdecls("w"),
                ),
            )
            self.cell._tc.get_or_add_tcPr().append(shading)  # noqa: SLF001



"""Module for extending python-docx Run objects."""

import dataclasses

from docx import shared
from docx.text import paragraph as docx_paragraph


@dataclasses.dataclass
class FindRun:
    """Data class for maintaing find results in runs.

    Attributes:
        paragraph: The paragraph containing the text.
        run_indices: The run indices of the text needle's start and end.
        character_indices: The character indices of the text in the runs.
            The first index is the start of the text in the first run containing
            the text. The second index is the end of the text in the last run.
    """

    paragraph: docx_paragraph.Paragraph
    run_indices: tuple[int, int]
    character_indices: tuple[int, int]

    @property
    def runs(self) -> list[docx_paragraph.Run]:
        """Returns the runs containing the text."""
        return self.paragraph.runs[self.run_indices[0] : self.run_indices[1] + 1]

    def replace(self, replace: str) -> None:
        """Replaces the text in the runs with the replacement text.

        Args:
            replace: The text to replace.
        """
        start = self.character_indices[0]
        end = self.character_indices[1]
        if len(self.runs) > 1:
            self.runs[0].text = (
                self.runs[0].text[:start] + replace + self.runs[-1].text[end:]
            )
        else:
            self.runs[0].text = (
                self.runs[0].text[:start] + replace + self.runs[0].text[end:]
            )
        self.run_indices = (self.run_indices[0], self.run_indices[0])
        self.character_indices = (start, start + len(replace))

    def __lt__(self, other: "FindRun") -> bool:
        """Sorts FindRun in order of appearance in the paragraph.

        Makes FindRun objects sortable.

        Args:
            other: The other FindRun object.

        Returns:
            True if the character index of the first run is less than the
            character index of the other run.
        """
        if self.paragraph != other.paragraph:
            msg = "Cannot compare FindRun objects from different paragraphs."
            raise ValueError(msg)

        if self.run_indices[0] == other.run_indices[0]:
            return self.character_indices[0] < other.character_indices[0]
        return self.run_indices[0] < other.run_indices[0]


class ExtendRun:
    """Extends a python-docx Word run with additional functionality."""

    def __init__(self, run: docx_paragraph.Run) -> None:
        """Initializes an ExtendRun object.

        Args:
            run: The run to extend.
        """
        self.run = run

    def format(
        self,
        *,
        bold: bool | None = None,
        italics: bool | None = None,
        underline: bool | None = None,
        strike: bool | None = None,
        superscript: bool | None = None,
        subscript: bool | None = None,
        font_size: int | None = None,
        font_rgb: tuple[int, int, int] | None = None,
    ) -> None:
        """Formats a run in a Word document.

        Args:
            bold: Whether to bold the run.
            italics: Whether to italicize the run.
            underline: Whether to underline the run.
            strike: Whether to strike through the run.
            superscript: Whether to superscript the run.
            subscript: Whether to subscript the run.
            font_size: The font size of the run.
            font_rgb: The font color of the run.
        """
        if superscript and subscript:
            msg = "Cannot have superscript and subscript at the same time."
            raise ValueError(msg)

        if bold is not None:
            self.run.bold = bold
        if italics is not None:
            self.run.italic = italics
        if underline is not None:
            self.run.underline = underline
        if strike is not None:
            self.run.strike = strike
        if superscript is not None:
            self.run.font.superscript = superscript
        if subscript is not None:
            self.run.font.subscript = subscript
        if font_size is not None:
            self.run.font.size = font_size
        if font_rgb is not None:
            self.run.font.color.rgb = shared.RGBColor(*font_rgb)
